unit UMachine;

{$mode delphi}{$H+}

interface

uses
  Classes, SysUtils;

type
  T = Integer;

  TOpcode = (opNop, opLoad, opHalt, opLoadConst, opLoadMem, opStoreMem, opNeg,
    opJump, opJumpRel, opPush, opPop, opCall, opCallRel, opRet, opInput,
    opOutput, opInc, opDec);

  TOpcodeHandler = procedure of object;
  TInputEvent = function (Port: T): T of object;
  TOutputEvent = procedure (Port, Value: T) of object;

  { TCPU }

  TCPU = class(TComponent)
  private
    FOnInput: TInputEvent;
    FOnOutput: TOutputEvent;
    OpcodeHandlers: array[TOpcode] of TOpcodeHandler;
    function ReadNext: T; inline;
    procedure OpcodeHalt;
    procedure OpcodeNop;
    procedure OpcodeLoad;
    procedure OpcodeLoadConst;
    procedure OpcodeLoadMem;
    procedure OpcodeStoreMem;
    procedure OpcodeNeg;
    procedure OpcodeJump;
    procedure OpcodeJumpRel;
    procedure OpcodePush;
    procedure OpcodePop;
    procedure OpcodeCall;
    procedure OpcodeCallRel;
    procedure OpcodeReturn;
    procedure OpcodeOutput;
    procedure OpcodeInput;
    procedure OpcodeInc;
    procedure OpcodeDec;
  public
    Registers: array of T;
    IP: T;
    SP: T;
    Memory: array of T;
    Terminated: Boolean;
    procedure Start;
    procedure Stop;
    constructor Create(AOwner: TComponent); override;
  published
    property OnInput: TInputEvent read FOnInput write FOnInput;
    property OnOutput: TOutputEvent read FOnOutput write FOnOutput;
  end;


implementation

{ TCPU }

function TCPU.ReadNext: T;
begin
  Result := Memory[IP];
  Inc(IP);
end;

procedure TCPU.OpcodeHalt;
begin
  Terminated := True;
end;

procedure TCPU.OpcodeNop;
begin
  // Do nothing
end;

procedure TCPU.OpcodeLoad;
var
  P1: T;
  P2: T;
begin
  P1 := ReadNext;
  P2 := ReadNext;
  Registers[P1] := Registers[P2];
end;

procedure TCPU.OpcodeLoadConst;
var
  P1: T;
  P2: T;
begin
  P1 := ReadNext;
  P2 := ReadNext;
  Registers[P1] := P2;
end;

procedure TCPU.OpcodeLoadMem;
var
  P1: T;
  P2: T;
begin
  P1 := ReadNext;
  P2 := ReadNext;
  Registers[P1] := Memory[Registers[P2]];
end;

procedure TCPU.OpcodeStoreMem;
var
  P1: T;
  P2: T;
begin
  P1 := ReadNext;
  P2 := ReadNext;
  Memory[Registers[P1]] := Registers[P2];
end;

procedure TCPU.OpcodeNeg;
var
  P1: T;
begin
  P1 := ReadNext;
  Registers[P1] := -Registers[P1];
end;

procedure TCPU.OpcodeJump;
begin
  IP := ReadNext;
end;

procedure TCPU.OpcodeJumpRel;
begin
  IP := IP + ReadNext;
end;

procedure TCPU.OpcodePush;
begin
  Memory[SP] := Registers[ReadNext];
  Dec(SP);
end;

procedure TCPU.OpcodePop;
begin
  Inc(SP);
  Registers[ReadNext] := Memory[SP];
end;

procedure TCPU.OpcodeCall;
var
  Addr: T;
begin
  Addr := ReadNext;
  Memory[SP] := IP;
  Dec(SP);
  IP := Addr;
end;

procedure TCPU.OpcodeCallRel;
var
  Addr: T;
begin
  Addr := ReadNext;
  Memory[SP] := IP;
  Dec(SP);
  IP := IP + Addr;
end;

procedure TCPU.OpcodeReturn;
begin
  Inc(SP);
  IP := Memory[SP];
end;

procedure TCPU.OpcodeOutput;
var
  R1: T;
  R2: T;
begin
  R1 := ReadNext;
  R2 := ReadNext;
  if Assigned(FOnOutput) then
    FOnOutput(Registers[R1], Registers[R2]);
end;

procedure TCPU.OpcodeInput;
var
  R1: T;
  R2: T;
begin
  R1 := ReadNext;
  R2 := ReadNext;
  if Assigned(FOnInput) then
    Registers[R1] := FOnInput(Registers[R2]);
end;

procedure TCPU.OpcodeInc;
var
  R: T;
begin
  R := ReadNext;
  Registers[R] := Registers[R] + 1;
end;

procedure TCPU.OpcodeDec;
var
  R: T;
begin
  R := ReadNext;
  Registers[R] := Registers[R] - 1;
end;

procedure TCPU.Start;
var
  Opcode: T;
begin
  Terminated := False;
  IP := 0;
  SP := Length(Memory);
  while not Terminated do begin
    Opcode := ReadNext;
    if (Opcode >= 0) and (Opcode <= T(High(TOpcode))) then
      OpcodeHandlers[TOpcode(Opcode)]
      else raise Exception.Create(Format('Unsupported instruction %d', [Opcode]));
  end;
end;

procedure TCPU.Stop;
begin
  Terminated := True;
end;

constructor TCPU.Create(AOwner: TComponent);
begin
  inherited;
  SetLength(Registers, 16);
  SetLength(Memory, 1024);
  OpcodeHandlers[opNop] := OpcodeNop;
  OpcodeHandlers[opLoad] := OpcodeLoad;
  OpcodeHandlers[opHalt] := OpcodeHalt;
  OpcodeHandlers[opLoadConst] := OpcodeLoadConst;
  OpcodeHandlers[opLoadMem] := OpcodeLoadMem;
  OpcodeHandlers[opStoreMem] := OpcodeStoreMem;
  OpcodeHandlers[opNeg] := OpcodeNeg;
  OpcodeHandlers[opJump] := OpcodeJump;
  OpcodeHandlers[opJumpRel] := OpcodeJumpRel;
  OpcodeHandlers[opPush] := OpcodePush;
  OpcodeHandlers[opPop] := OpcodePop;
  OpcodeHandlers[opCall] := OpcodeCall;
  OpcodeHandlers[opCallRel] := OpcodeCallRel;
  OpcodeHandlers[opRet] := OpcodeReturn;
  OpcodeHandlers[opInput] := OpcodeInput;
  OpcodeHandlers[opOutput] := OpcodeOutput;
  OpcodeHandlers[opInc] := OpcodeInc;
  OpcodeHandlers[opDec] := OpcodeDec;
end;

end.

